@@ -96,7 +96,7 @@ pt_to_px(double pt, id_type os_window_id) {
 
 
 OSWindow*
-current_os_window() {
+current_os_window(void) {
     if (global_state.callback_os_window) return global_state.callback_os_window;
     for (size_t i = 0; i < global_state.num_os_windows; i++) {
         if (global_state.os_windows[i].is_focused) return global_state.os_windows + i;
@@ -104,6 +104,38 @@ current_os_window() {
     return global_state.os_windows;
 }
 
+static id_type
+last_focused_os_window_id(void) {
+    id_type ans = 0, max_fc_count = 0;
+    for (size_t i = 0; i < global_state.num_os_windows; i++) {
+        OSWindow *w = &global_state.os_windows[i];
+        if (w->last_focused_counter > max_fc_count) {
+            ans = w->id; max_fc_count = w->last_focused_counter;
+        }
+    }
+    return ans;
+}
+
+
+static id_type
+current_focused_os_window_id(void) {
+    for (size_t i = 0; i < global_state.num_os_windows; i++) {
+        OSWindow *w = &global_state.os_windows[i];
+        if (w->is_focused) { return w->id; }
+    }
+    return 0;
+}
+
+
+OSWindow*
+os_window_for_id(id_type os_window_id) {
+    for (size_t i = 0; i < global_state.num_os_windows; i++) {
+        OSWindow *w = global_state.os_windows + i;
+        if (w->id == os_window_id) return w;
+    }
+    return NULL;
+}
+
 OSWindow*
 os_window_for_kitty_window(id_type kitty_window_id) {
     for (size_t i = 0; i < global_state.num_os_windows; i++) {
@@ -134,15 +166,16 @@ window_for_window_id(id_type kitty_window_id) {
 
 static void
 send_bgimage_to_gpu(BackgroundImageLayout layout, BackgroundImage *bgimage) {
-    RepeatStrategy r;
+    RepeatStrategy r = REPEAT_DEFAULT;
     switch (layout) {
         case SCALED:
         case CLAMPED:
+        case CENTER_CLAMPED:
+        case CENTER_SCALED:
             r = REPEAT_CLAMP; break;
         case MIRRORED:
             r = REPEAT_MIRROR; break;
         case TILING:
-        default:
             r = REPEAT_DEFAULT; break;
     }
     bgimage->texture_id = 0;
@@ -165,15 +198,15 @@ free_bgimage(BackgroundImage **bgimage, bool release_texture) {
 }
 
 OSWindow*
-add_os_window() {
+add_os_window(void) {
     WITH_OS_WINDOW_REFS
     ensure_space_for(&global_state, os_windows, OSWindow, global_state.num_os_windows + 1, capacity, 1, true);
     OSWindow *ans = global_state.os_windows + global_state.num_os_windows++;
     zero_at_ptr(ans);
     ans->id = ++global_state.os_window_id_counter;
     ans->tab_bar_render_data.vao_idx = create_cell_vao();
-    ans->gvao_idx = create_graphics_vao();
     ans->background_opacity = OPT(background_opacity);
+    ans->created_at = monotonic();
 
     bool wants_bg = OPT(background_image) && OPT(background_image)[0] != 0;
     if (wants_bg) {
@@ -213,22 +246,19 @@ add_tab(id_type os_window_id) {
 static void
 create_gpu_resources_for_window(Window *w) {
     w->render_data.vao_idx = create_cell_vao();
-    w->render_data.gvao_idx = create_graphics_vao();
 }
 
 static void
 release_gpu_resources_for_window(Window *w) {
     if (w->render_data.vao_idx > -1) remove_vao(w->render_data.vao_idx);
     w->render_data.vao_idx = -1;
-    if (w->render_data.gvao_idx > -1) remove_vao(w->render_data.gvao_idx);
-    w->render_data.gvao_idx = -1;
 }
 
 static bool
-set_window_logo(Window *w, const char *path, const ImageAnchorPosition pos, float alpha, bool is_default) {
+set_window_logo(Window *w, const char *path, const ImageAnchorPosition pos, float alpha, bool is_default, char *png_data, size_t png_data_size) {
     bool ok = false;
     if (path && path[0]) {
-        window_logo_id_t wl = find_or_create_window_logo(global_state.all_window_logos, path);
+        window_logo_id_t wl = find_or_create_window_logo(global_state.all_window_logos, path, png_data, png_data_size);
         if (wl) {
             if (w->window_logo.id) decref_window_logo(global_state.all_window_logos, w->window_logo.id);
             w->window_logo.id = wl;
@@ -254,14 +284,13 @@ initialize_window(Window *w, PyObject *title, bool init_gpu_resources) {
     w->visible = true;
     w->title = title;
     Py_XINCREF(title);
-    if (!set_window_logo(w, OPT(default_window_logo), OPT(window_logo_position), OPT(window_logo_alpha), true)) {
+    if (!set_window_logo(w, OPT(default_window_logo), OPT(window_logo_position), OPT(window_logo_alpha), true, NULL, 0)) {
         log_error("Failed to load default window logo: %s", OPT(default_window_logo));
         if (PyErr_Occurred()) PyErr_Print();
     }
     if (init_gpu_resources) create_gpu_resources_for_window(w);
     else {
         w->render_data.vao_idx = -1;
-        w->render_data.gvao_idx = -1;
     }
 }
 
@@ -313,6 +342,8 @@ destroy_window(Window *w) {
     Py_CLEAR(w->render_data.screen); Py_CLEAR(w->title);
     Py_CLEAR(w->title_bar_data.last_drawn_title_object_id);
     free(w->title_bar_data.buf); w->title_bar_data.buf = NULL;
+    Py_CLEAR(w->url_target_bar_data.last_drawn_title_object_id);
+    free(w->url_target_bar_data.buf); w->url_target_bar_data.buf = NULL;
     release_gpu_resources_for_window(w);
     if (w->window_logo.id) {
         decref_window_logo(global_state.all_window_logos, w->window_logo.id);
@@ -444,10 +475,7 @@ destroy_os_window_item(OSWindow *w) {
         remove_tab_inner(w, tab->id);
     }
     Py_CLEAR(w->window_title); Py_CLEAR(w->tab_bar_render_data.screen);
-    if (w->offscreen_texture_id) free_texture(&w->offscreen_texture_id);
-    if (w->offscreen_framebuffer) free_framebuffer(&w->offscreen_framebuffer);
     remove_vao(w->tab_bar_render_data.vao_idx);
-    remove_vao(w->gvao_idx);
     free(w->tabs); w->tabs = NULL;
     free_bgimage(&w->bgimage, true);
     w->bgimage = NULL;
@@ -490,6 +518,7 @@ set_active_window(id_type os_window_id, id_type tab_id, id_type window_id) {
         (void)window;
         tab->active_window = w;
         osw->needs_render = true;
+        set_os_window_chrome(osw);
     END_WITH_WINDOW;
 }
 
@@ -602,7 +631,7 @@ send_pending_click_to_window_id(id_type timer_id UNUSED, void *data) {
 }
 
 bool
-update_ime_position_for_window(id_type window_id, bool force, bool lost_focus) {
+update_ime_position_for_window(id_type window_id, bool force, int update_focus) {
     for (size_t o = 0; o < global_state.num_os_windows; o++) {
         OSWindow *osw = global_state.os_windows + o;
         for (size_t t = 0; t < osw->num_tabs; t++) {
@@ -610,11 +639,14 @@ update_ime_position_for_window(id_type window_id, bool force, bool lost_focus) {
             for (size_t w = 0; w < qtab->num_windows; w++) {
                 Window *window = qtab->windows + w;
                 if (window->id == window_id) {
-                    if (window->render_data.screen && (force || osw->is_focused)) {
+                    // The screen may not be ready after the new window is created and focused, and still needs to enable IME.
+                    if ((window->render_data.screen && (force || osw->is_focused)) || update_focus > 0) {
                         OSWindow *orig = global_state.callback_os_window;
                         global_state.callback_os_window = osw;
-                        if (lost_focus) update_ime_focus(osw, false);
-                        else update_ime_position(window, window->render_data.screen);
+                        if (update_focus) update_ime_focus(osw, update_focus > 0);
+                        if (update_focus >= 0 && window->render_data.screen) {
+                            update_ime_position(window, window->render_data.screen);
+                        }
                         global_state.callback_os_window = orig;
                         return true;
                     }
@@ -653,9 +685,9 @@ update_ime_position_for_window(id_type window_id, bool force, bool lost_focus) {
 PYWRAP1(update_ime_position_for_window) {
     id_type window_id;
     int force = 0;
-    int lost_focus = 0;
-    PA("K|pp", &window_id, &force, &lost_focus);
-    if (update_ime_position_for_window(window_id, force, lost_focus)) Py_RETURN_TRUE;
+    int update_focus = 0;
+    PA("K|pi", &window_id, &force, &update_focus);
+    if (update_ime_position_for_window(window_id, force, update_focus)) Py_RETURN_TRUE;
     Py_RETURN_FALSE;
 }
 
@@ -663,6 +695,15 @@ PYWRAP0(next_window_id) {
     return PyLong_FromUnsignedLongLong(global_state.window_id_counter + 1);
 }
 
+PYWRAP0(last_focused_os_window_id) {
+    return PyLong_FromUnsignedLongLong(last_focused_os_window_id());
+}
+
+PYWRAP0(current_focused_os_window_id) {
+    return PyLong_FromUnsignedLongLong(current_focused_os_window_id());
+}
+
+
 PYWRAP1(handle_for_window_id) {
     id_type os_window_id;
     PA("K", &os_window_id);
@@ -819,22 +860,31 @@ PYWRAP0(current_application_quit_request) {
 PYWRAP1(focus_os_window) {
     id_type os_window_id;
     int also_raise = 1;
-    PA("K|p", &os_window_id, &also_raise);
+    const char *activation_token = NULL;
+    PA("K|pz", &os_window_id, &also_raise, &activation_token);
     WITH_OS_WINDOW(os_window_id)
-        if (!os_window->is_focused) focus_os_window(os_window, also_raise);
+        if (!os_window->is_focused || (activation_token && activation_token[0])) focus_os_window(os_window, also_raise, activation_token);
         Py_RETURN_TRUE;
     END_WITH_OS_WINDOW
     Py_RETURN_FALSE;
 }
 
-PYWRAP1(set_titlebar_color) {
+PYWRAP1(run_with_activation_token) {
+    for (size_t o = 0; o < global_state.num_os_windows; o++) {
+        OSWindow *os_window = global_state.os_windows + o;
+        if (os_window->is_focused) {
+            run_with_activation_token_in_os_window(os_window, args);
+            break;
+        }
+    }
+    Py_RETURN_NONE;
+}
+
+PYWRAP1(set_os_window_chrome) {
     id_type os_window_id;
-    unsigned int color;
-    int use_system_color = 0;
-    unsigned int system_color = 0;
-    PA("KI|pI", &os_window_id, &color, &use_system_color, &system_color);
+    PA("K", &os_window_id);
     WITH_OS_WINDOW(os_window_id)
-        set_titlebar_color(os_window, color, use_system_color, system_color);
+        set_os_window_chrome(os_window);
         Py_RETURN_TRUE;
     END_WITH_OS_WINDOW
     Py_RETURN_FALSE;
@@ -889,7 +939,6 @@ PYWRAP1(set_window_render_data) {
     WITH_WINDOW(os_window_id, tab_id, window_id);
         Py_CLEAR(window->render_data.screen);
         d.vao_idx = window->render_data.vao_idx;
-        d.gvao_idx = window->render_data.gvao_idx;
         init_screen_render_data(osw, &g, &d);
         window->render_data = d;
         window->geometry = g;
@@ -924,12 +973,15 @@ PYWRAP1(sync_os_window_title) {
 
 PYWRAP1(set_os_window_title) {
     id_type os_window_id;
-    const char *title;
-    PA("Ks", &os_window_id, &title);
+    PyObject *title;
+    PA("KU", &os_window_id, &title);
     WITH_OS_WINDOW(os_window_id)
-        if (strlen(title)) {
+        if (PyUnicode_GetLength(title)) {
             os_window->title_is_overriden = true;
-            set_os_window_title(os_window, title);
+            Py_XDECREF(os_window->window_title);
+            os_window->window_title = title;
+            Py_INCREF(title);
+            set_os_window_title(os_window, PyUnicode_AsUTF8(title));
         } else {
             os_window->title_is_overriden = false;
             if (os_window->window_title) set_os_window_title(os_window, PyUnicode_AsUTF8(os_window->window_title));
@@ -939,6 +991,15 @@ PYWRAP1(set_os_window_title) {
     Py_RETURN_NONE;
 }
 
+PYWRAP1(get_os_window_title) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id)
+        if (os_window->window_title) return Py_BuildValue("O", os_window->window_title);
+    END_WITH_OS_WINDOW
+    Py_RETURN_NONE;
+}
+
 
 
 PYWRAP1(pt_to_px) {
@@ -1037,7 +1098,7 @@ PYWRAP0(apply_options_update) {
             for (size_t w = 0; w < tab->num_windows; w++) {
                 Window *window = tab->windows + w;
                 if (window->window_logo.using_default) {
-                    set_window_logo(window, OPT(default_window_logo), OPT(window_logo_position), OPT(window_logo_alpha), true);
+                    set_window_logo(window, OPT(default_window_logo), OPT(window_logo_position), OPT(window_logo_alpha), true, NULL, 0);
                 }
             }
         }
@@ -1066,20 +1127,45 @@ PYWRAP1(patch_global_colors) {
     Py_RETURN_NONE;
 }
 
+PYWRAP1(update_tab_bar_edge_colors) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id)
+        if (os_window->tab_bar_render_data.screen) {
+            if (get_line_edge_colors(os_window->tab_bar_render_data.screen, &os_window->tab_bar_edge_color.left, &os_window->tab_bar_edge_color.right)) { Py_RETURN_TRUE; }
+        }
+    END_WITH_OS_WINDOW
+    Py_RETURN_FALSE;
+}
+
 static PyObject*
 pyset_background_image(PyObject *self UNUSED, PyObject *args) {
     const char *path;
     PyObject *layout_name = NULL;
     PyObject *os_window_ids;
     int configured = 0;
-    PA("zO!|pO", &path, &PyTuple_Type, &os_window_ids, &configured, &layout_name);
+    char *png_data = NULL; Py_ssize_t png_data_size = 0;
+    PA("zO!|pOy#", &path, &PyTuple_Type, &os_window_ids, &configured, &layout_name, &png_data, &png_data_size);
     size_t size;
     BackgroundImageLayout layout = PyUnicode_Check(layout_name) ? bglayout(layout_name) : OPT(background_image_layout);
     BackgroundImage *bgimage = NULL;
     if (path) {
         bgimage = calloc(1, sizeof(BackgroundImage));
         if (!bgimage) return PyErr_NoMemory();
-        if (!png_path_to_bitmap(path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size)) {
+        bool ok;
+        if (png_data) {
+            FILE *fp = fmemopen(png_data, png_data_size, "r");
+            if (fp == NULL) {
+                PyErr_SetFromErrnoWithFilename(PyExc_OSError, path);
+                free(bgimage);
+                return NULL;
+            }
+            ok = png_from_file_pointer(fp, path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size);
+            fclose(fp);
+        } else {
+            ok = png_path_to_bitmap(path, &bgimage->bitmap, &bgimage->width, &bgimage->height, &size);
+        }
+        if (!ok) {
             PyErr_Format(PyExc_ValueError, "Failed to load image from: %s", path);
             free(bgimage);
             return NULL;
@@ -1113,6 +1199,11 @@ PYWRAP0(destroy_global_data) {
     Py_RETURN_NONE;
 }
 
+PYWRAP0(wakeup_main_loop) {
+    wakeup_main_loop();
+    Py_RETURN_NONE;
+}
+
 static void
 destroy_mock_window(PyObject *capsule) {
     Window *w = PyCapsule_GetPointer(capsule, "Window");
@@ -1166,6 +1257,18 @@ move_cursor_to_mouse_if_in_prompt(id_type os_window_id, id_type tab_id, id_type
     return moved;
 }
 
+static PyObject*
+pyupdate_pointer_shape(PyObject *self UNUSED, PyObject *args) {
+    id_type os_window_id;
+    PA("K", &os_window_id);
+    WITH_OS_WINDOW(os_window_id);
+    OSWindow *orig = global_state.callback_os_window;
+    global_state.callback_os_window = os_window;
+    update_mouse_pointer_shape();
+    global_state.callback_os_window = orig;
+    END_WITH_OS_WINDOW;
+    Py_RETURN_NONE;
+}
 
 static PyObject*
 pymouse_selection(PyObject *self UNUSED, PyObject *args) {
@@ -1182,10 +1285,11 @@ PYWRAP1(set_window_logo) {
     id_type os_window_id, tab_id, window_id;
     const char *path; PyObject *position;
     float alpha = 0.5;
-    PA("KKKsUf", &os_window_id, &tab_id, &window_id, &path, &position, &alpha);
+    char *png_data = NULL; Py_ssize_t png_data_size = 0;
+    PA("KKKsUf|y#", &os_window_id, &tab_id, &window_id, &path, &position, &alpha, &png_data, &png_data_size);
     bool ok = false;
     WITH_WINDOW(os_window_id, tab_id, window_id);
-    ok = set_window_logo(window, path, PyObject_IsTrue(position) ? bganchor(position) : OPT(window_logo_position), (0 <= alpha && alpha <= 1) ? alpha : OPT(window_logo_alpha), false);
+    ok = set_window_logo(window, path, PyObject_IsTrue(position) ? bganchor(position) : OPT(window_logo_position), (0 <= alpha && alpha <= 1) ? alpha : OPT(window_logo_alpha), false, png_data, png_data_size);
     END_WITH_WINDOW;
     if (ok) Py_RETURN_TRUE;
     Py_RETURN_FALSE;
@@ -1219,7 +1323,6 @@ THREE_ID_OBJ(update_window_title)
 THREE_ID(remove_window)
 THREE_ID(detach_window)
 THREE_ID(attach_window)
-PYWRAP1(resolve_key_mods) { int mods, kitty_mod; PA("ii", &kitty_mod, &mods); return PyLong_FromLong(resolve_mods(kitty_mod, mods)); }
 PYWRAP1(add_tab) { return PyLong_FromUnsignedLongLong(add_tab(PyLong_AsUnsignedLongLong(args))); }
 PYWRAP1(add_window) { PyObject *title; id_type a, b; PA("KKO", &a, &b, &title); return PyLong_FromUnsignedLongLong(add_window(a, b, title)); }
 PYWRAP0(current_os_window) { OSWindow *w = current_os_window(); if (!w) Py_RETURN_NONE; return PyLong_FromUnsignedLongLong(w->id); }
@@ -1234,8 +1337,11 @@ KK5I(add_borders_rect)
 #define MW(name, arg_type) {#name, (PyCFunction)py##name, arg_type, NULL}
 
 static PyMethodDef module_methods[] = {
+    MW(update_pointer_shape, METH_VARARGS),
     MW(current_os_window, METH_NOARGS),
     MW(next_window_id, METH_NOARGS),
+    MW(last_focused_os_window_id, METH_NOARGS),
+    MW(current_focused_os_window_id, METH_NOARGS),
     MW(set_options, METH_VARARGS),
     MW(get_options, METH_NOARGS),
     MW(click_mouse_url, METH_VARARGS),
@@ -1245,7 +1351,6 @@ static PyMethodDef module_methods[] = {
     MW(mouse_selection, METH_VARARGS),
     MW(set_window_logo, METH_VARARGS),
     MW(set_in_sequence_mode, METH_O),
-    MW(resolve_key_mods, METH_VARARGS),
     MW(handle_for_window_id, METH_VARARGS),
     MW(update_ime_position_for_window, METH_VARARGS),
     MW(pt_to_px, METH_VARARGS),
@@ -1270,25 +1375,29 @@ static PyMethodDef module_methods[] = {
     MW(mark_os_window_for_close, METH_VARARGS),
     MW(set_application_quit_request, METH_VARARGS),
     MW(current_application_quit_request, METH_NOARGS),
-    MW(set_titlebar_color, METH_VARARGS),
+    MW(set_os_window_chrome, METH_VARARGS),
     MW(focus_os_window, METH_VARARGS),
     MW(mark_tab_bar_dirty, METH_O),
+    MW(run_with_activation_token, METH_O),
     MW(change_background_opacity, METH_VARARGS),
     MW(background_opacity_of, METH_O),
     MW(update_window_visibility, METH_VARARGS),
     MW(sync_os_window_title, METH_VARARGS),
+    MW(get_os_window_title, METH_VARARGS),
     MW(set_os_window_title, METH_VARARGS),
     MW(global_font_size, METH_VARARGS),
     MW(set_background_image, METH_VARARGS),
     MW(os_window_font_size, METH_VARARGS),
     MW(set_os_window_size, METH_VARARGS),
     MW(get_os_window_size, METH_VARARGS),
+    MW(update_tab_bar_edge_colors, METH_VARARGS),
     MW(set_boss, METH_O),
     MW(get_boss, METH_NOARGS),
     MW(apply_options_update, METH_NOARGS),
     MW(patch_global_colors, METH_VARARGS),
     MW(create_mock_window, METH_VARARGS),
     MW(destroy_global_data, METH_NOARGS),
+    MW(wakeup_main_loop, METH_NOARGS),
 
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
@@ -1301,7 +1410,7 @@ finalize(void) {
     if (detached_windows.windows) free(detached_windows.windows);
     detached_windows.capacity = 0;
 #define F(x) free(OPT(x)); OPT(x) = NULL;
-    F(background_image); F(bell_path); F(default_window_logo);
+    F(background_image); F(bell_path); F(bell_theme); F(default_window_logo);
 #undef F
     // we leak the texture here since it is not guaranteed
     // that freeing the texture will work during shutdown and
@@ -1310,9 +1419,8 @@ finalize(void) {
     free_bgimage(&global_state.bgimage, false);
     free_window_logo_table(&global_state.all_window_logos);
     global_state.bgimage = NULL;
-    free_url_prefixes();
-    free(OPT(select_by_word_characters)); OPT(select_by_word_characters) = NULL;
-    free(OPT(url_excluded_characters)); OPT(url_excluded_characters) = NULL;
+
+    free_allocs_in_options(&global_state.opts);
 }
 
 bool
@@ -1333,6 +1441,10 @@ init_state(PyObject *module) {
     PyModule_AddIntConstant(module, "IMPERATIVE_CLOSE_REQUESTED", IMPERATIVE_CLOSE_REQUESTED);
     PyModule_AddIntConstant(module, "NO_CLOSE_REQUESTED", NO_CLOSE_REQUESTED);
     PyModule_AddIntConstant(module, "CLOSE_BEING_CONFIRMED", CLOSE_BEING_CONFIRMED);
+    PyModule_AddIntMacro(module, WINDOW_NORMAL);
+    PyModule_AddIntMacro(module, WINDOW_FULLSCREEN);
+    PyModule_AddIntMacro(module, WINDOW_MAXIMIZED);
+    PyModule_AddIntMacro(module, WINDOW_MINIMIZED);
     register_at_exit_cleanup_func(STATE_CLEANUP_FUNC, finalize);
     return true;
 }
