@@ -2,15 +2,25 @@
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
+import sys
 import tempfile
 
 from kitty.config import build_ansi_color_table, defaults
 from kitty.fast_data_types import (
-    ColorProfile, Cursor as C, HistoryBuf, LineBuf, Color,
-    parse_input_from_terminal, truncate_point_for_length, wcswidth, wcwidth, strip_csi
+    Color,
+    ColorProfile,
+    HistoryBuf,
+    LineBuf,
+    expand_ansi_c_escapes,
+    parse_input_from_terminal,
+    strip_csi,
+    truncate_point_for_length,
+    wcswidth,
+    wcwidth,
 )
+from kitty.fast_data_types import Cursor as C
 from kitty.rgb import to_color
-from kitty.utils import is_path_in_temp_dir, sanitize_title
+from kitty.utils import is_ok_to_read_image_file, is_path_in_temp_dir, sanitize_title, sanitize_url_for_dispay_to_user
 
 from . import BaseTest, filled_cursor, filled_history_buf, filled_line_buf
 
@@ -18,11 +28,10 @@ from . import BaseTest, filled_cursor, filled_history_buf, filled_line_buf
 def create_lbuf(*lines):
     maxw = max(map(len, lines))
     ans = LineBuf(len(lines), maxw)
-    prev_full_length = False
     for i, l0 in enumerate(lines):
         ans.line(i).set_text(l0, 0, len(l0), C())
-        ans.set_continued(i, prev_full_length)
-        prev_full_length = len(l0) == maxw
+        if i > 0:
+            ans.set_continued(i, len(lines[i-1]) == maxw)
     return ans
 
 
@@ -71,9 +80,9 @@ class TestDataTypes(BaseTest):
                 self.assertFalse(c.reverse)
                 self.assertTrue(c.bold)
         self.assertFalse(old.is_continued(0))
-        old.set_continued(0, True)
-        self.assertTrue(old.is_continued(0))
-        self.assertFalse(old.is_continued(1))
+        old.set_continued(1, True)
+        self.assertTrue(old.is_continued(1))
+        self.assertFalse(old.is_continued(0))
 
         lb = filled_line_buf(5, 5, filled_cursor())
         lb2 = LineBuf(5, 5)
@@ -182,9 +191,11 @@ class TestDataTypes(BaseTest):
         l0.add_combining_char(0, '\U000e0100')
         self.ae(l0[0], ' \u0300\U000e0100')
         l0.add_combining_char(0, '\u0302')
-        self.ae(l0[0], ' \u0300\u0302')
+        self.ae(l0[0], ' \u0300\U000e0100\u0302')
+        l0.add_combining_char(0, '\u0301')
+        self.ae(l0[0], ' \u0300\U000e0100\u0301')
         self.ae(l0[1], '\0')
-        self.ae(str(l0), ' \u0300\u0302')
+        self.ae(str(l0), ' \u0300\U000e0100\u0301')
         t = 'Testing with simple text'
         lb = LineBuf(2, len(t))
         l0 = lb.line(0)
@@ -394,9 +405,10 @@ class TestDataTypes(BaseTest):
         self.ae(tpl('a\U0001f337', 2), 1)
         self.ae(tpl('a\U0001f337', 3), 2)
         self.ae(tpl('a\U0001f337b', 4), 3)
-        self.ae(sanitize_title('a\0\01 \t\n\f\rb'), 'a b')
         self.ae(tpl('a\x1b[31mbc', 2), 7)
 
+        self.ae(sanitize_title('a\0\01 \t\n\f\rb'), 'a b')
+
         def tp(*data, leftover='', text='', csi='', apc='', ibp=False):
             text_r, csi_r, apc_r, rest = [], [], [], []
             left = ''
@@ -432,13 +444,30 @@ class TestDataTypes(BaseTest):
                 self.assertTrue(is_path_in_temp_dir(os.path.join(prefix, path)))
         for path in ('/home/xy/d.png', '/tmp/../home/x.jpg'):
             self.assertFalse(is_path_in_temp_dir(os.path.join(path)))
+        for path in ('/proc/self/cmdline', os.devnull):
+            if os.path.exists(path):
+                with open(path) as pf:
+                    self.assertFalse(is_ok_to_read_image_file(path, pf.fileno()), path)
+        fifo = os.path.join(tempfile.gettempdir(), 'test-kitty-fifo')
+        os.mkfifo(fifo)
+        fifo_fd = os.open(fifo, os.O_RDONLY | os.O_NONBLOCK)
+        try:
+            self.assertFalse(is_ok_to_read_image_file(fifo, fifo_fd), fifo)
+        finally:
+            os.close(fifo_fd)
+            os.remove(fifo)
+        if os.path.isdir('/dev/shm'):
+            with tempfile.NamedTemporaryFile(dir='/dev/shm') as tf:
+                self.assertTrue(is_ok_to_read_image_file(tf.name, tf.fileno()), fifo)
+        self.ae(sanitize_url_for_dispay_to_user(
+            'h://a\u0430b.com/El%20Ni%C3%B1o/'), 'h://xn--ab-7kc.com/El Niño/')
 
     def test_color_profile(self):
         c = ColorProfile()
         c.update_ansi_color_table(build_ansi_color_table())
         for i in range(8):
             col = getattr(defaults, f'color{i}')
-            self.assertEqual(c.as_color(i << 8 | 1), col)
+            self.ae(c.as_color(i << 8 | 1), col)
         self.ae(c.as_color(255 << 8 | 1), Color(0xee, 0xee, 0xee))
 
     def test_historybuf(self):
@@ -513,7 +542,7 @@ class TestDataTypes(BaseTest):
         self.ae(l2.as_ansi(), '\x1b[1;2;3;7;9;34;48:2:1:2:3;58:5:5m' '1'
                 '\x1b[22;23;27;29;39;49;59m' '0000')
         lb = filled_line_buf()
-        for i in range(lb.ynum):
+        for i in range(1, lb.ynum + 1):
             lb.set_continued(i, True)
         a = []
         lb.as_ansi(a.append)
@@ -530,3 +559,56 @@ class TestDataTypes(BaseTest):
         q('a\x1bbc', 'ac')
         q('a\x1b[bc', 'ac')
         q('a\x1b[12;34:43mbc', 'abc')
+
+    def test_single_key(self):
+        from kitty.fast_data_types import GLFW_MOD_KITTY, GLFW_MOD_SHIFT, SingleKey
+        for m in (GLFW_MOD_KITTY, GLFW_MOD_SHIFT):
+            s = SingleKey(mods=m)
+            self.ae(s.mods, m)
+        self.ae(tuple(iter(SingleKey())), (0, False, 0))
+        self.ae(tuple(SingleKey(key=sys.maxunicode, mods=GLFW_MOD_SHIFT, is_native=True)), (GLFW_MOD_SHIFT, True, sys.maxunicode))
+        self.ae(repr(SingleKey()), 'SingleKey()')
+        self.ae(repr(SingleKey(key=23, mods=2, is_native=True)), 'SingleKey(mods=2, is_native=True, key=23)')
+        self.ae(repr(SingleKey(key=23, mods=2)), 'SingleKey(mods=2, key=23)')
+        self.ae(repr(SingleKey(key=23)), 'SingleKey(key=23)')
+        self.ae(repr(SingleKey(key=0x1008ff57)), 'SingleKey(key=269025111)')
+        self.ae(repr(SingleKey(key=23)._replace(mods=2)), 'SingleKey(mods=2, key=23)')
+        self.ae(repr(SingleKey(key=23)._replace(key=-1, mods=GLFW_MOD_KITTY)), f'SingleKey(mods={GLFW_MOD_KITTY})')
+        self.assertEqual(SingleKey(key=1), SingleKey(key=1))
+        self.assertEqual(hash(SingleKey(key=1)), hash(SingleKey(key=1)))
+        self.assertNotEqual(hash(SingleKey(key=1, mods=2)), hash(SingleKey(key=1)))
+        self.assertNotEqual(SingleKey(key=1, mods=2), SingleKey(key=1))
+
+    def test_notify_identifier_sanitization(self):
+        from kitty.notify import sanitize_identifier_pat
+        self.ae(sanitize_identifier_pat().sub('', '\x1b\nabc\n[*'), 'abc')
+
+    def test_bracketed_paste_sanitizer(self):
+        from kitty.utils import sanitize_for_bracketed_paste
+        for x in ('\x1b[201~ab\x9b201~cd', '\x1b[201\x1b[201~~ab'):  # ]]]
+            q = sanitize_for_bracketed_paste(x.encode('utf-8'))
+            self.assertNotIn(b'\x1b[201~', q)
+            self.assertNotIn('\x9b201~'.encode('utf-8'), q)
+            self.assertIn(b'ab', q)
+
+    def test_expand_ansi_c_escapes(self):
+        for src, expected in {
+            'abc': 'abc',
+            r'a\ab': 'a\ab',
+            r'a\eb': 'a\x1bb',
+            r'a\r\nb': 'a\r\nb',
+            r'a\c b': 'a\0b',
+            r'a\c': 'a\\c',
+            r'a\x1bb': 'a\x1bb',
+            r'a\x1b': 'a\x1b',
+            r'a\x1': 'a\x01',
+            r'a\x1g': 'a\x01g',
+            r'a\z\"': 'a\\z"',
+            r'a\123b': 'a\123b',
+            r'a\128b': 'a\0128b',
+            r'a\u1234e': 'a\u1234e',
+            r'a\U1f1eez': 'a\U0001f1eez',
+            r'a\x1\\':    "a\x01\\",
+        }.items():
+            actual = expand_ansi_c_escapes(src)
+            self.ae(expected, actual)
