@@ -1,5 +1,4 @@
 #!/usr/bin/env python3
-# vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
 import time
@@ -7,11 +6,8 @@ from base64 import standard_b64encode
 from binascii import hexlify
 from functools import partial
 
-from kitty.fast_data_types import CURSOR_BLOCK, parse_bytes, parse_bytes_dump
-from kitty.notify import (
-    NotificationCommand, handle_notification_cmd, notification_activated,
-    reset_registry
-)
+from kitty.fast_data_types import CURSOR_BLOCK, base64_decode, base64_encode, parse_bytes, parse_bytes_dump
+from kitty.notify import NotificationCommand, handle_notification_cmd, notification_activated, reset_registry
 
 from . import BaseTest
 
@@ -45,6 +41,16 @@ class TestParser(BaseTest):
             q.append(('draw', current))
         self.ae(tuple(q), cmds)
 
+    def test_base64(self):
+        for src, expected in {
+            'bGlnaHQgdw==': 'light w',
+            'bGlnaHQgd28=': 'light wo',
+            'bGlnaHQgd29y': 'light wor',
+        }.items():
+            self.ae(base64_decode(src.encode()), expected.encode(), f'Decoding of {src} failed')
+            self.ae(base64_decode(src.replace('=', '').encode()), expected.encode(), f'Decoding of {src} failed')
+            self.ae(base64_encode(expected.encode()), src.replace('=', '').encode(), f'Encoding of {expected} failed')
+
     def test_simple_parsing(self):
         s = self.create_screen()
         pb = partial(self.parse_bytes_dump, s)
@@ -100,10 +106,13 @@ class TestParser(BaseTest):
         s.cursor_back(5)
         pb('x\033[2@y', 'x', ('screen_insert_characters', 2), 'y')
         self.ae(str(s.line(0)), 'xy bc')
-        pb('x\033[2;7@y', 'x', ('screen_insert_characters', 2), 'y')
+        pb('x\033[2;7@y', 'x', ('CSI code @ has 2 > 1 parameters',), 'y')
+        pb('x\033[2;-7@y', 'x', ('CSI code @ has 2 > 1 parameters',), 'y')
+        pb('x\033[-2@y', 'x', ('CSI code @ is not allowed to have negative parameter (-2)',), 'y')
+        pb('x\033[2-3@y', 'x', ('CSI code can contain hyphens only at the start of numbers',), 'y')
         pb('x\033[@y', 'x', ('screen_insert_characters', 1), 'y')
         pb('x\033[345@y', 'x', ('screen_insert_characters', 345), 'y')
-        pb('x\033[345;@y', 'x', ('CSI code 0x40 has unsupported start modifier: 0x0 or end modifier: 0x3b',), 'y')
+        pb('x\033[345;@y', 'x', ('screen_insert_characters', 345), 'y')
         pb('\033[H', ('screen_cursor_position', 1, 1))
         self.ae(s.cursor.x, 0), self.ae(s.cursor.y, 0)
         pb('\033[4H', ('screen_cursor_position', 4, 1))
@@ -123,7 +132,7 @@ class TestParser(BaseTest):
         s.reset()
 
         def sgr(params):
-            return (('select_graphic_rendition', '{} '.format(x)) for x in params.split())
+            return (('select_graphic_rendition', f'{x} ') for x in params.split())
 
         pb('\033[1;2;3;4;7;9;34;44m', *sgr('1 2 3 4 7 9 34 44'))
         for attr in 'bold italic reverse strikethrough dim'.split():
@@ -181,6 +190,9 @@ class TestParser(BaseTest):
         pb('\033[3 @', ('Shift left escape code not implemented',))
         pb('\033[3 A', ('Shift right escape code not implemented',))
         pb('\033[3;4 S', ('Select presentation directions escape code not implemented',))
+        pb('\033[1T', ('screen_reverse_scroll', 1))
+        pb('\033[T', ('screen_reverse_scroll', 1))
+        pb('\033[+T', ('screen_reverse_scroll_and_fill_from_scrollback', 1))
 
     def test_csi_code_rep(self):
         s = self.create_screen(8)
@@ -208,18 +220,18 @@ class TestParser(BaseTest):
         c = s.callbacks
         pb('a\033]2;x\\ryz\x9cbcde', 'a', ('set_title', 'x\\ryz'), 'bcde')
         self.ae(str(s.line(0)), 'abcde')
-        self.ae(c.titlebuf, 'x\\ryz')
+        self.ae(c.titlebuf, ['x\\ryz'])
         c.clear()
         pb('\033]\x07', ('set_title', ''), ('set_icon', ''))
-        self.ae(c.titlebuf, ''), self.ae(c.iconbuf, '')
+        self.ae(c.titlebuf, ['']), self.ae(c.iconbuf, '')
         pb('\033]ab\x07', ('set_title', 'ab'), ('set_icon', 'ab'))
-        self.ae(c.titlebuf, 'ab'), self.ae(c.iconbuf, 'ab')
+        self.ae(c.titlebuf, ['', 'ab']), self.ae(c.iconbuf, 'ab')
         c.clear()
         pb('\033]2;;;;\x07', ('set_title', ';;;'))
-        self.ae(c.titlebuf, ';;;')
+        self.ae(c.titlebuf, [';;;'])
         c.clear()
         pb('\033]2;\x07', ('set_title', ''))
-        self.ae(c.titlebuf, '')
+        self.ae(c.titlebuf, [''])
         pb('\033]110\x07', ('set_dynamic_color', 110, ''))
         self.ae(c.colorbuf, '')
         c.clear()
@@ -233,6 +245,11 @@ class TestParser(BaseTest):
         pb('\033]8;moo\x07', ('Ignoring malformed OSC 8 code',))
         pb('\033]8;id=xyz;\x07', ('set_active_hyperlink', 'xyz', None))
         pb('\033]8;moo:x=z:id=xyz:id=abc;http://yay;.com\x07', ('set_active_hyperlink', 'xyz', 'http://yay;.com'))
+        c.clear()
+        payload = '1' * 1024
+        pb(f'\033]52;p;{payload}\x07', ('clipboard_control', 52, f'p;{payload}'))
+        c.clear()
+        pb('\033]52;p;xyz\x07', ('clipboard_control', 52, 'p;xyz'))
 
     def test_desktop_notify(self):
         reset_registry()
@@ -302,7 +319,7 @@ class TestParser(BaseTest):
         c = s.callbacks
         pb = partial(self.parse_bytes_dump, s)
         q = hexlify(b'kind').decode('ascii')
-        pb('a\033P+q{}\x9cbcde'.format(q), 'a', ('screen_request_capabilities', 43, q), 'bcde')
+        pb(f'a\033P+q{q}\x9cbcde', 'a', ('screen_request_capabilities', 43, q), 'bcde')
         self.ae(str(s.line(0)), 'abcde')
         self.ae(c.wtcbuf, '1+r{}={}'.format(q, '1b5b313b3242').encode('ascii'))
         c.clear()
@@ -314,12 +331,12 @@ class TestParser(BaseTest):
         for sgr in '0;34;102;1;2;3;4 0;38:5:200;58:2:10:11:12'.split():
             expected = set(sgr.split(';')) - {'0'}
             c.clear()
-            parse_bytes(s, '\033[{}m\033P$qm\033\\'.format(sgr).encode('ascii'))
+            parse_bytes(s, f'\033[{sgr}m\033P$qm\033\\'.encode('ascii'))
             r = c.wtcbuf.decode('ascii').partition('r')[2].partition('m')[0]
             self.ae(expected, set(r.split(';')))
         c.clear()
         pb('\033P$qr\033\\', ('screen_request_capabilities', ord('$'), 'r'))
-        self.ae(c.wtcbuf, '\033P1$r{};{}r\033\\'.format(s.margin_top + 1, s.margin_bottom + 1).encode('ascii'))
+        self.ae(c.wtcbuf, f'\033P1$r{s.margin_top + 1};{s.margin_bottom + 1}r\033\\'.encode('ascii'))
 
     def test_sc81t(self):
         s = self.create_screen()
@@ -337,10 +354,11 @@ class TestParser(BaseTest):
         timeout = 0.1
         s.set_pending_timeout(timeout)
         pb = partial(self.parse_bytes_dump, s)
+
         pb('\033P=1s\033\\', ('screen_start_pending_mode',))
         pb('a')
         self.ae(str(s.line(0)), '')
-        pb('\033P=2s\033\\', ('screen_stop_pending_mode',), ('draw', 'a'))
+        pb('\033P=2s\033\\', ('draw', 'a'), ('screen_stop_pending_mode',))
         self.ae(str(s.line(0)), 'a')
         pb('\033P=1s\033\\', ('screen_start_pending_mode',))
         pb('b')
@@ -349,22 +367,45 @@ class TestParser(BaseTest):
         pb('c', ('draw', 'bc'))
         self.ae(str(s.line(0)), 'abc')
         pb('\033P=1s\033\\d', ('screen_start_pending_mode',))
-        pb('\033P=2s\033\\', ('screen_stop_pending_mode',), ('draw', 'd'))
+        pb('\033P=2s\033\\', ('draw', 'd'), ('screen_stop_pending_mode',))
         pb('\033P=1s\033\\e', ('screen_start_pending_mode',))
         pb('\033P'), pb('='), pb('2s')
-        pb('\033\\', ('screen_stop_pending_mode',), ('draw', 'e'))
-        pb('\033P=1sxyz;.;\033\\''\033P=2skjf".,><?_+)98\033\\', ('screen_start_pending_mode',), ('screen_stop_pending_mode',))
+        pb('\033\\', ('draw', 'e'), ('screen_stop_pending_mode',))
+        pb('\033P=1sxyz;.;\033\\''\033P=2skjf".,><?_+)98\033\\', ('screen_start_pending_mode',))
         pb('\033P=1s\033\\f\033P=1s\033\\', ('screen_start_pending_mode',), ('screen_start_pending_mode',))
+        pb('\033P=2s\033\\', ('draw', 'f'), ('screen_stop_pending_mode',))
+        pb('\033P=1s\033\\XXX\033P=2s\033\\', ('screen_start_pending_mode',), ('draw', 'XXX'), ('screen_stop_pending_mode',))
+
+        pb('\033[?2026hXXX\033[?2026l', ('screen_set_mode', 2026, 1), ('draw', 'XXX'), ('screen_reset_mode', 2026, 1))
+        pb('\033[?2026h\033[32ma\033[?2026l', ('screen_set_mode', 2026, 1), ('select_graphic_rendition', '32 '), ('draw', 'a'), ('screen_reset_mode', 2026, 1))
+        pb('\033[?2026h\033P+q544e\033\\ama\033P=2s\033\\',
+           ('screen_set_mode', 2026, 1), ('screen_request_capabilities', 43, '544e'), ('draw', 'ama'), ('screen_stop_pending_mode',))
+        pb('\033P=1s\033\\\033(B\033P=2s\033\\', ('screen_start_pending_mode',), ('screen_designate_charset', 0, 66), ('screen_stop_pending_mode',))
+
+        s.reset()
+        s.set_pending_timeout(timeout)
+        pb('\033[?2026h', ('screen_set_mode', 2026, 1),)
+        pb('\033P+q')
+        time.sleep(1.2 * timeout)
+        pb(
+            '544e' + '\033\\\033P=2s\033\\',
+            ('screen_request_capabilities', 43, '544e'),
+            ('Pending mode stop command issued while not in pending mode, this can be '
+             'either a bug in the terminal application or caused by a timeout with no '
+             'data received for too long or by too much data in pending mode',),
+            ('screen_stop_pending_mode',)
+        )
+        self.assertEqual(str(s.line(0)), '')
 
     def test_oth_codes(self):
         s = self.create_screen()
         pb = partial(self.parse_bytes_dump, s)
         for prefix in '\033_', '\u009f':
             for suffix in '\u009c', '\033\\':
-                pb('a{}+\\++{}bcde'.format(prefix, suffix), ('draw', 'a'), ('Unrecognized APC code: 0x2b',), ('draw', 'bcde'))
+                pb(f'a{prefix}+\\++{suffix}bcde', ('draw', 'a'), ('Unrecognized APC code: 0x2b',), ('draw', 'bcde'))
         for prefix in '\033^', '\u009e':
             for suffix in '\u009c', '\033\\':
-                pb('a{}+\\++{}bcde'.format(prefix, suffix), ('draw', 'a'), ('Unrecognized PM code: 0x2b',), ('draw', 'bcde'))
+                pb(f'a{prefix}+\\++{suffix}bcde', ('draw', 'a'), ('Unrecognized PM code: 0x2b',), ('draw', 'bcde'))
 
     def test_graphics_command(self):
         from base64 import standard_b64encode
@@ -378,8 +419,10 @@ class TestParser(BaseTest):
                     k[p] = v.encode('ascii')
             for f in 'action delete_action transmission_type compressed'.split():
                 k.setdefault(f, b'\0')
-            for f in ('format more id data_sz data_offset width height x_offset y_offset data_height data_width'
-                      ' num_cells num_lines cell_x_offset cell_y_offset z_index placement_id image_number quiet').split():
+            for f in ('format more id data_sz data_offset width height x_offset y_offset data_height data_width cursor_movement'
+                      ' num_cells num_lines cell_x_offset cell_y_offset z_index placement_id image_number quiet unicode_placement'
+                      ' parent_id parent_placement_id offset_from_parent_x offset_from_parent_y'
+            ).split():
                 k.setdefault(f, 0)
             p = k.pop('payload', '').encode('utf-8')
             k['payload_sz'] = len(p)
@@ -389,7 +432,7 @@ class TestParser(BaseTest):
             pb('\033_G{};{}\033\\'.format(cmd, enc(kw.get('payload', ''))), c(**kw))
 
         def e(cmd, err):
-            pb('\033_G{}\033\\'.format(cmd), (err,))
+            pb(f'\033_G{cmd}\033\\', (err,))
 
         s = self.create_screen()
         pb = partial(self.parse_bytes_dump, s)
