@@ -1,9 +1,7 @@
 #!/usr/bin/env python3
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
-from kitty.fast_data_types import (
-    DECAWM, DECCOLM, DECOM, IRM, Cursor, parse_bytes
-)
+from kitty.fast_data_types import DECAWM, DECCOLM, DECOM, IRM, Cursor, parse_bytes
 from kitty.marks import marker_from_function, marker_from_regex
 from kitty.window import pagerhist
 
@@ -210,22 +208,23 @@ class TestScreen(BaseTest):
             s.reset_dirty()
             s.cursor.x, s.cursor.y = 2, 1
             s.cursor.bold = True
+            self.ae(continuations(s), (True, True, True, True, False))
 
         def all_lines(s):
             return tuple(str(s.line(i)) for i in range(s.lines))
 
         def continuations(s):
-            return tuple(s.line(i).is_continued() for i in range(s.lines))
+            return tuple(s.line(i).last_char_has_wrapped_flag() for i in range(s.lines))
 
         init()
         s.erase_in_display(0)
         self.ae(all_lines(s), ('12345', '12', '', '', ''))
-        self.ae(continuations(s), (False, True, False, False, False))
+        self.ae(continuations(s), (True, False, False, False, False))
 
         init()
         s.erase_in_display(1)
         self.ae(all_lines(s), ('', '   45', '12345', '12345', '12345'))
-        self.ae(continuations(s), (False, False, True, True, True))
+        self.ae(continuations(s), (False, True, True, True, False))
 
         init()
         s.erase_in_display(2)
@@ -293,6 +292,18 @@ class TestScreen(BaseTest):
         self.ae(str(s.linebuf), '55555\n66666\n77777\n88888\n99999')
         s.resize(5, 2)
         self.ae(str(s.linebuf), '88\n88\n99\n99\n9')
+        s = self.create_screen()
+        s.draw('a' * s.columns)
+        s.linefeed(), s.carriage_return()
+        s.draw('bb')
+        s.resize(s.lines, s.columns - 2)
+        self.ae(str(s.linebuf), 'aaa\naa\nbb\n\n')
+        s.cursor.y = s.cursor.x = 0
+        s.draw('x' * len(str(s.line(0))))
+        s.linefeed(), s.carriage_return()
+        s.draw('x' * len(str(s.line(1))))
+        s.resize(s.lines, s.columns + 4)
+        self.ae(str(s.linebuf), 'xxx\nxx\nbb\n\n')
 
     def test_cursor_after_resize(self):
 
@@ -320,6 +331,15 @@ class TestScreen(BaseTest):
         s.resize(s.lines - 1, s.columns)
         self.ae(x_before, s.cursor.x)
 
+        s = self.create_screen()
+        s.draw('abc')
+        b = s.cursor.x
+        s.resize(7, s.columns)
+        self.assertEqual(s.cursor.x, b)
+        s.cursor.x = 0
+        s.resize(5, s.columns)
+        self.assertEqual(s.cursor.x, 0)
+
     def test_scrollback_fill_after_resize(self):
         def prepare_screen(content=()):
             ans = self.create_screen(options={'scrollback_fill_enlarged_window': True})
@@ -526,16 +546,20 @@ class TestScreen(BaseTest):
             s.draw(str(i) * s.columns)
         s.start_selection(0, 0)
         s.update_selection(4, 4)
-        expected = ('55555', '\n66666', '\n77777', '\n88888', '\n99999')
-        self.ae(s.text_for_selection(), expected)
+
+        def ts(*args):
+            return ''.join(s.text_for_selection(*args))
+
+        expected = ''.join(('55555', '\n66666', '\n77777', '\n88888', '\n99999'))
+        self.ae(ts(), expected)
         s.scroll(2, True)
-        self.ae(s.text_for_selection(), expected)
+        self.ae(ts(), expected)
         s.reset()
         s.draw('ab   cd')
         s.start_selection(0, 0)
         s.update_selection(1, 3)
-        self.ae(s.text_for_selection(), ('ab   ', 'cd'))
-        self.ae(s.text_for_selection(False, True), ('ab', 'cd'))
+        self.ae(ts(), ''.join(('ab   ', 'cd')))
+        self.ae(ts(False, True), ''.join(('ab', 'cd')))
         s.reset()
         s.draw('ab        cd')
         s.start_selection(0, 0)
@@ -609,6 +633,38 @@ class TestScreen(BaseTest):
         s.draw('bcdef')
         self.ae(as_text(s, True), '\x1b[ma\x1b]8;;moo\x1b\\bcde\x1b[mf\n\n\n\x1b]8;;\x1b\\')
 
+    def test_wrapping_serialization(self):
+        from kitty.window import as_text
+        s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
+        s.draw('aabbccddeeff')
+        self.ae(as_text(s, add_history=True), 'aabbccddeeff')
+        self.assertNotIn('\n', as_text(s, add_history=True, as_ansi=True))
+
+        s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
+        s.draw('1'), s.carriage_return(), s.linefeed()
+        s.draw('2'), s.carriage_return(), s.linefeed()
+        s.draw('3'), s.carriage_return(), s.linefeed()
+        s.draw('4'), s.carriage_return(), s.linefeed()
+        s.draw('5'), s.carriage_return(), s.linefeed()
+        s.draw('6'), s.carriage_return(), s.linefeed()
+        s.draw('7')
+        self.ae(as_text(s, add_history=True), '1\n2\n3\n4\n5\n6\n7')
+
+        s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
+        s.draw('aabb')
+        s.cursor.y = 0
+        s.carriage_return(), s.linefeed()
+        self.ae(as_text(s, add_history=True), 'aabb')
+
+        s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
+        s.draw('a'), s.carriage_return(), s.linefeed()
+        s.cursor.y = 0
+        s.draw('aabb')
+        self.ae(as_text(s), 'aabb')
+        s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
+        s.draw('a😀')
+        self.ae(as_text(s), 'a😀')
+
     def test_pagerhist(self):
         hsz = 8
         s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': hsz})
@@ -645,17 +701,17 @@ class TestScreen(BaseTest):
         s = self.create_screen(options={'scrollback_pager_history_size': 2048})
         text = '\x1b[msoft\r\x1b[mbreak\nnext😼cat'
         w(text)
-        self.ae(contents(), text + '\n')
+        self.ae(contents(), text)
         s.historybuf.pagerhist_rewrap(2)
-        self.ae(contents(), '\x1b[mso\rft\x1b[m\rbr\rea\rk\nne\rxt\r😼\rca\rt\n')
+        self.ae(contents(), '\x1b[mso\rft\x1b[m\rbr\rea\rk\nne\rxt\r😼\rca\rt')
 
         s = self.create_screen(options={'scrollback_pager_history_size': 8})
         w('😼')
-        self.ae(contents(), '😼\n')
+        self.ae(contents(), '😼')
         w('abcd')
-        self.ae(contents(), '😼abcd\n')
+        self.ae(contents(), '😼abcd')
         w('e')
-        self.ae(contents(), 'abcde\n')
+        self.ae(contents(), 'abcde')
 
     def test_user_marking(self):
 
@@ -951,6 +1007,7 @@ class TestScreen(BaseTest):
             s.draw(before + url + after)
             ae(url, x=x + 1 + len(before), y=y)
 
+
         t('http://moo.com')
         t('http://moo.com/something?else=+&what-')
         for (st, e) in '() {} [] <>'.split():
@@ -960,6 +1017,7 @@ class TestScreen(BaseTest):
         for trailer in '{([':
             t('http://moo.com', after=trailer)
         t('http://moo.com', x=s.columns - 9)
+        t('https://wraps-by-one-char.com', before='[', after=']')
 
     def test_prompt_marking(self):
         s = self.create_screen()
@@ -997,9 +1055,9 @@ class TestScreen(BaseTest):
         self.assertTrue(s.scroll_to_prompt())
         self.ae(str(s.visual_line(0)), '$ 1')
 
-        def lco(as_ansi=False):
+        def lco(as_ansi=False, which=0):
             a = []
-            if s.cmd_output(0, a.append, as_ansi):
+            if s.cmd_output(which, a.append, as_ansi):
                 pht = pagerhist(s, as_ansi=as_ansi, upto_output_start=True)
                 if pht:
                     a.insert(0, pht)
@@ -1019,7 +1077,7 @@ class TestScreen(BaseTest):
         s.draw('abcd'), s.index(), s.carriage_return()
         s.draw('12'), s.index(), s.carriage_return()
         self.ae(fco(), '')
-        self.ae(lco(), 'abcd\n12')
+        self.ae(lco(), 'abcd\n12\n')
         s = self.create_screen()
         mark_prompt(), s.draw('$ 0')
         s.carriage_return(), s.index()
@@ -1049,12 +1107,12 @@ class TestScreen(BaseTest):
         # last cmd output
         # test: find upwards
         self.ae(s.scrolled_by, 0)
-        self.ae(lco(), '0x\n1x')
+        self.ae(lco(), '0x\n1x\n')
         # get output after scroll up
         s.scroll_to_prompt()
         self.ae(s.scrolled_by, 4)
         self.ae(str(s.visual_line(0)), '$ 0')
-        self.ae(lco(), '0x\n1x')
+        self.ae(lco(), '0x\n1x\n')
 
         # first cmd output on screen
         # test: find around
@@ -1062,11 +1120,11 @@ class TestScreen(BaseTest):
         s.scroll(2, False)
         self.ae(s.scrolled_by, 2)
         self.ae(str(s.visual_line(0)), '1')
-        self.ae(fco(), '0x\n1x')
+        self.ae(fco(), '0x\n1x\n')
         # test: find downwards
         s.scroll(2, False)
         self.ae(str(s.visual_line(0)), '$ 1')
-        self.ae(fco(), '0x\n1x')
+        self.ae(fco(), '0x\n1x\n')
 
         # resize
         # get last cmd output with continued output mark
@@ -1074,15 +1132,85 @@ class TestScreen(BaseTest):
         s.resize(4, 5)
         s.scroll_to_prompt(-4)
         self.ae(str(s.visual_line(0)), '$ 0')
-        self.ae(lco(), '0long_line\n0l\n1l')
+        self.ae(lco(), '0long_line\n0l\n1l\n')
 
         # last visited cmd output
         self.ae(lvco(), '0\n1\n2')
         s.scroll_to_prompt(1)
         self.ae(lvco(), '0x\n1x')
 
+        # last command output without line break
+        s = self.create_screen(cols=10, lines=3)
+        draw_prompt('p1')
+        mark_output(), s.draw('running')
+        self.ae(lco(), 'running')
+        s.index(), s.carriage_return()
+        self.ae(lco(), 'running\n')
+
         # last command output from pager history
         s = self.create_screen()
         draw_prompt('p1')
         draw_output(30)
         self.ae(tuple(map(int, lco().split())), tuple(range(0, 30)))
+
+        # last non empty command output
+        s = self.create_screen()
+        draw_prompt('a'), draw_output(2, 'a')
+        draw_prompt('b'), mark_output()
+        self.ae(lco(), '')
+        self.ae(lco(which=3), '0a\n1a')
+        s.draw('running'), s.index(),  s.carriage_return()
+        self.ae(lco(which=3), 'running\n')
+        s = self.create_screen()
+        draw_prompt('p1')
+        draw_output(30)
+        self.ae(tuple(map(int, lco(which=3).split())), tuple(range(0, 30)))
+        s = self.create_screen()
+        draw_prompt('p1')
+        draw_output(2, 'a')
+        draw_prompt('p1')
+        draw_prompt('p1')
+        self.ae(lco(which=3), '0a\n1a')
+
+    def test_pointer_shapes(self):
+        from kitty.window import set_pointer_shape
+        s = self.create_screen()
+        c = s.callbacks
+        response = ''
+
+        def cb(data):
+            nonlocal response
+            response = set_pointer_shape(s, data)
+        c.set_pointer_shape = cb
+
+        def send(a):
+            nonlocal response
+            response = ''
+            parse_bytes(s, f'\x1b]22;{a}\x1b\\'.encode())
+            return response
+
+        self.ae(send('?__current__'), '0')
+        self.ae(send('?__default__,__grabbed__,default,ne-resize,crosshair,XXX'), 'text,default,1,1,1,0')
+
+        def t(q, e=None):
+            self.ae(send(q), '')
+            self.ae(send('?__current__'), e)
+
+        t('default', 'default')
+        s.reset()
+        self.ae(send('?__current__'), '0')
+        t('=crosshair', 'crosshair')
+        t('<', '0')
+        t('=crosshair', 'crosshair')
+        t('', '0')
+        t('>help', 'help')
+        t('>wait', 'wait')
+        t('<', 'help')
+        t('<', '0')
+        t('default,help', 'help')
+        t('<', '0')
+        t('>default,help', 'help')
+        t('<', 'default')
+        t('<', '0')
+        t('=left_ptr', 'default')
+        t('=fleur', 'move')
