@@ -11,11 +11,10 @@ import unittest
 from contextlib import contextmanager
 from functools import lru_cache, partial
 
-from kitty.constants import kitty_base_dir, terminfo_dir
+from kitty.bash import decode_ansi_c_quoted_string
+from kitty.constants import kitten_exe, kitty_base_dir, shell_integration_dir, terminfo_dir
 from kitty.fast_data_types import CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE
-from kitty.shell_integration import (
-    setup_bash_env, setup_fish_env, setup_zsh_env
-)
+from kitty.shell_integration import setup_bash_env, setup_fish_env, setup_zsh_env
 
 from . import BaseTest
 
@@ -25,10 +24,11 @@ def bash_ok():
     v = shutil.which('bash')
     if not v:
         return False
-    o = subprocess.check_output([v, '-c', 'echo "${BASH_VERSION}"']).decode('utf-8').strip()
-    if not o or int(o[0]) < 5:
+    o = subprocess.check_output([v, '-c', 'echo "${BASH_VERSINFO[0]}\n${BASH_VERSINFO[4]}"']).decode('utf-8').splitlines()
+    if not o:
         return False
-    return True
+    major_ver, relstatus = o[0], o[-1]
+    return int(major_ver) >= 5 and relstatus == 'release'
 
 
 def basic_shell_env(home_dir):
@@ -40,6 +40,7 @@ def basic_shell_env(home_dir):
         'KITTY_SHELL_INTEGRATION': 'enabled',
         'KITTY_INSTALLATION_DIR': kitty_base_dir,
         'BASH_SILENCE_DEPRECATION_WARNING': '1',
+        'PYTHONDONTWRITEBYTECODE': '1',
     }
     for x in ('USER', 'LANG'):
         if os.environ.get(x):
@@ -47,13 +48,12 @@ def basic_shell_env(home_dir):
     return ans
 
 
-def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh'):
+def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh', with_kitten=False):
     ans = basic_shell_env(home_dir)
     if shell == 'zsh':
         argv.insert(1, '--noglobalrcs')
-        ans['ZLE_RPROMPT_INDENT'] = '0'
         with open(os.path.join(home_dir, '.zshrc'), 'w') as f:
-            print(rc + '\n', file=f)
+            print(rc + '\nZLE_RPROMPT_INDENT=0', file=f)
         setup_zsh_env(ans, argv)
     elif shell == 'fish':
         conf_dir = os.path.join(home_dir, '.config', 'fish')
@@ -64,10 +64,14 @@ def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh'):
             print(rc + '\n', file=f)
         setup_fish_env(ans, argv)
     elif shell == 'bash':
-        setup_bash_env(ans, argv)
-        ans['KITTY_BASH_INJECT'] += ' posix'
-        ans['KITTY_BASH_POSIX_ENV'] = os.path.join(home_dir, '.bashrc')
-        with open(ans['KITTY_BASH_POSIX_ENV'], 'w') as f:
+        bashrc = os.path.join(home_dir, '.bashrc')
+        if with_kitten:
+            ans['KITTY_RUNNING_BASH_INTEGRATION_TEST'] = bashrc
+        else:
+            setup_bash_env(ans, argv)
+            ans['KITTY_BASH_INJECT'] += ' posix'
+            ans['KITTY_BASH_POSIX_ENV'] = bashrc
+        with open(bashrc, 'w') as f:
             # ensure LINES and COLUMNS are kept up to date
             print('shopt -s checkwinsize', file=f)
             if rc:
@@ -77,13 +81,17 @@ def safe_env_for_running_shell(argv, home_dir, rc='', shell='zsh'):
 
 class ShellIntegration(BaseTest):
 
+    with_kitten = False
+
     @contextmanager
     def run_shell(self, shell='zsh', rc='', cmd='', setup_env=None):
-        home_dir = os.path.realpath(tempfile.mkdtemp())
+        home_dir = self.home_dir = os.path.realpath(tempfile.mkdtemp())
         cmd = cmd or shell
         cmd = shlex.split(cmd.format(**locals()))
-        env = (setup_env or safe_env_for_running_shell)(cmd, home_dir, rc=rc, shell=shell)
+        env = (setup_env or safe_env_for_running_shell)(cmd, home_dir, rc=rc, shell=shell, with_kitten=self.with_kitten)
         try:
+            if self.with_kitten:
+                cmd = [kitten_exe(), 'run-shell', '--shell', shlex.join(cmd)]
             pty = self.create_pty(cmd, cwd=home_dir, env=env)
             i = 10
             while i > 0 and not pty.screen_contents().strip():
@@ -107,7 +115,7 @@ RPS1="{rps1}"
                 pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
             except TimeoutError as e:
                 raise AssertionError(f'Cursor was not changed to beam. Screen contents: {repr(pty.screen_contents())}') from e
-            self.ae(pty.screen_contents(), q)
+            pty.wait_till(lambda: pty.screen_contents() == q)
             self.ae(pty.callbacks.titlebuf[-1], '~')
             pty.callbacks.clear()
             pty.send_cmd_to_child('mkdir test && ls -a')
@@ -143,7 +151,18 @@ RPS1="{rps1}"
             pty.callbacks.clear()
             pty.send_cmd_to_child('printf "%s\x16\a%s" "a" "b"')
             pty.wait_till(lambda: 'ab' in pty.screen_contents())
+            self.assertTrue(pty.screen.last_reported_cwd.endswith(self.home_dir))
             self.assertIn('%s^G%s', pty.screen_contents())
+            q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
+            os.mkdir(q)
+            pty.send_cmd_to_child(f'cd {q}')
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(q))
+        with self.run_shell(rc=f'''PS1="{ps1}"\nexport ES="a\n b c\nd"''') as pty:
+            pty.callbacks.clear()
+            pty.send_cmd_to_child('clone-in-kitty')
+            pty.wait_till(lambda: len(pty.callbacks.clone_cmds) == 1)
+            env = pty.callbacks.clone_cmds[0].env
+            self.ae(env.get('ES'), 'a\n b c\nd')
 
     @unittest.skipUnless(shutil.which('fish'), 'fish not installed')
     def test_fish_integration(self):
@@ -163,13 +182,23 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 1)
             self.ae(pty.screen_contents(), q)
 
-            # XDG_DATA_DIRS
-            pty.send_cmd_to_child('set -q XDG_DATA_DIRS; or echo ok')
-            pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 2)
-            self.ae(str(pty.screen.line(1)), 'ok')
+            # shell integration dir must no be in XDG_DATA_DIRS
+            pty.send_cmd_to_child(f'string match -q -- "*{shell_integration_dir}*" "$XDG_DATA_DIRS" || echo "XDD_OK"')
+            pty.wait_till(lambda: 'XDD_OK' in pty.screen_contents())
+
+            # CWD reporting
+            self.assertTrue(pty.screen.last_reported_cwd.endswith(self.home_dir))
+            q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
+            os.mkdir(q)
+            pty.send_cmd_to_child(f'cd {q}')
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(q))
+            pty.send_cmd_to_child('cd -')
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(self.home_dir))
 
             # completion and prompt marking
+            pty.wait_till(lambda: 'cd -' not in pty.screen_contents().splitlines()[-1])
             pty.send_cmd_to_child('clear')
+            pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 1)
             pty.send_cmd_to_child('_test_comp_path')
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 2)
             q = '\n'.join(str(pty.screen.line(i)) for i in range(1, pty.screen.cursor.y))
@@ -203,7 +232,7 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.write_to_child('\x04')
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
             pty.send_cmd_to_child('_set_key vi')
-            pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BLOCK)
+            pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 3)
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
             pty.write_to_child('\x1b')
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BLOCK)
@@ -214,20 +243,12 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.write_to_child('i')
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
             pty.send_cmd_to_child('_set_key default')
-
-            # pipestatus
-            pty.send_cmd_to_child('clear;false|true|false')
-            pty.send_cmd_to_child('echo $pipestatus $status')
-            pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 2)
-            self.ae(pty.last_cmd_output(), '1 0 1 1')
-            pty.send_cmd_to_child('_set_status_prompt')
-            pty.send_cmd_to_child('false|true|false')
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 4)
-            self.assertTrue(str(pty.screen.line(pty.screen.cursor.y)).startswith(f'1 0 1 1 {fish_prompt}'))
+            pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
 
             pty.send_cmd_to_child('exit')
 
-    @unittest.skipUnless(bash_ok(), 'bash not installed or too old')
+    @unittest.skipUnless(bash_ok(), 'bash not installed, too old, or debug build')
     def test_bash_integration(self):
         ps1 = 'prompt> '
         with self.run_shell(
@@ -273,13 +294,22 @@ PS1="{ps1}"
             pty.write_to_child('\x04')
             pty.send_cmd_to_child('clear')
             pty.wait_till(lambda: pty.callbacks.titlebuf)
+        with self.run_shell(shell='bash', rc=f'''PS1="{ps1}"\ndeclare LOCAL_KSI_VAR=1''') as pty:
+            pty.callbacks.clear()
+            pty.send_cmd_to_child('declare')
+            pty.wait_till(lambda: 'LOCAL_KSI_VAR' in pty.screen_contents())
         with self.run_shell(shell='bash', rc=f'''PS1="{ps1}"''') as pty:
             pty.callbacks.clear()
             pty.send_cmd_to_child('printf "%s\x16\a%s" "a" "b"')
             pty.wait_till(lambda: pty.screen_contents().count(ps1) == 2)
             self.ae(pty.screen_contents(), f'{ps1}printf "%s^G%s" "a" "b"\nab{ps1}')
+            self.assertTrue(pty.screen.last_reported_cwd.endswith(self.home_dir))
             pty.send_cmd_to_child('echo $HISTFILE')
             pty.wait_till(lambda: '.bash_history' in pty.screen_contents())
+            q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
+            os.mkdir(q)
+            pty.send_cmd_to_child(f'cd {q}')
+            pty.wait_till(lambda: pty.screen.last_reported_cwd.endswith(q))
 
         for ps1 in ('line1\\nline\\2\\prompt> ', 'line1\nprompt> ', 'line1\\nprompt> ',):
             with self.subTest(ps1=ps1), self.run_shell(
@@ -303,21 +333,29 @@ PS1="{ps1}"
 
         # test startup file sourcing
 
-        def setup_env(excluded, argv, home_dir, rc='', shell='bash'):
+        def setup_env(excluded, argv, home_dir, rc='', shell='bash', with_kitten=self.with_kitten):
             ans = basic_shell_env(home_dir)
-            setup_bash_env(ans, argv)
+            if not with_kitten:
+                setup_bash_env(ans, argv)
             for x in {'profile', 'bash.bashrc', '.bash_profile', '.bash_login', '.profile', '.bashrc', 'rcfile'} - excluded:
                 with open(os.path.join(home_dir, x), 'w') as f:
-                    print(f'echo {x}', file=f)
+                    if x == '.bashrc' and rc:
+                        print(rc, file=f)
+                    else:
+                        print(f'echo [{x}]', file=f)
             ans['KITTY_BASH_ETC_LOCATION'] = home_dir
+            ans['PS1'] = 'PROMPT $ '
             return ans
 
-        def run_test(argv, *expected, excluded=()):
-            with self.subTest(argv=argv), self.run_shell(shell='bash', setup_env=partial(setup_env, set(excluded)), cmd=argv) as pty:
-                pty.wait_till(lambda: '$' in pty.screen_contents())
+        def run_test(argv, *expected, excluded=(), rc='', wait_string='PROMPT $', assert_not_in=False):
+            with self.subTest(argv=argv), self.run_shell(shell='bash', setup_env=partial(setup_env, set(excluded)), cmd=argv, rc=rc) as pty:
+                pty.wait_till(lambda: wait_string in pty.screen_contents())
                 q = pty.screen_contents()
                 for x in expected:
-                    self.assertIn(x, q)
+                    if assert_not_in:
+                        self.assertNotIn(f'[{x}]', q)
+                    else:
+                        self.assertIn(f'[{x}]', q)
 
         run_test('bash', 'bash.bashrc', '.bashrc')
         run_test('bash --rcfile rcfile', 'bash.bashrc', 'rcfile')
@@ -327,3 +365,32 @@ PS1="{ps1}"
         run_test('bash --noprofile -l')
         run_test('bash -l', 'profile', '.bash_login', excluded=('.bash_profile',))
         run_test('bash -l', 'profile', '.profile', excluded=('.bash_profile', '.bash_login'))
+
+        # test argument parsing and non-interactive shell
+        run_test('bash -s arg1 --rcfile rcfile', 'rcfile', rc='echo ok;read', wait_string='ok', assert_not_in=True)
+        run_test('bash +O login_shell -ic "echo ok;read"', 'bash.bashrc', excluded=('.bash_profile'), wait_string='ok', assert_not_in=True)
+        run_test('bash -l .bashrc', 'profile', rc='echo ok;read', wait_string='ok', assert_not_in=True)
+        run_test('bash -il -- .bashrc', 'profile', rc='echo ok;read', wait_string='ok')
+
+        with self.run_shell(shell='bash', setup_env=partial(setup_env, set()), cmd='bash',
+                            rc=f'''PS1="{ps1}"\nexport ES=$'a\n `b` c\n$d'\nexport ES2="XXX" ''') as pty:
+            pty.callbacks.clear()
+            pty.send_cmd_to_child('clone-in-kitty')
+            pty.wait_till(lambda: len(pty.callbacks.clone_cmds) == 1)
+            env = pty.callbacks.clone_cmds[0].env
+            self.ae(env.get('ES'), 'a\n `b` c\n$d', f'Screen contents: {pty.screen_contents()!r}')
+            self.ae(env.get('ES2'), 'XXX', f'Screen contents: {pty.screen_contents()!r}')
+        for q, e in {
+            'a': 'a',
+            r'a\ab': 'a\ab',
+            r'a\x7z': 'a\x07z',
+            r'a\7b': 'a\007b',
+            r'a\U1f345x': 'a🍅x',
+            r'a\c b': 'a\0b',
+        }.items():
+            q = q + "'"
+            self.ae(decode_ansi_c_quoted_string(q, 0)[0], e, f'Failed to decode: {q!r}')
+
+
+class ShellIntegrationWithKitten(ShellIntegration):
+    with_kitten = True
